/*********************************************************************
 *                                                                   *
 *                     XGS : Apple IIGS Emulator                     *
 *                                                                   *
 *        Written and Copyright (C)1996 by Joshua M. Thompson        *
 *                                                                   *
 *  You are free to distribute this code for non-commercial purposes *
 * I ask only that you notify me of any changes you make to the code *
 *     Commercial use is prohibited without my written permission    *
 *                                                                   *
 *********************************************************************/

/*
 * File: main.c
 *
 * Memory manager initialization and core routines.
 */

#include <xgs.h>
#include "memory.h"
#include "../disks/disks.h"
#include "../video/video.h"

int MEM_init(void)
{
	int	i,banks;
	FILE	*fp;

	if (mem_ramsize < 1) mem_ramsize = 1;
	if (mem_ramsize > 8) mem_ramsize = 8;
	banks = mem_ramsize * 16;

	printf("\nInitializing emulator memory\n");
	for (i = 0x02 ; i < 0xFC ; i++) memory[i] = NULL;

	printf("    - Allocating 256 KB for ROM (4 banks): ");
	for (i = 0xFC ; i <= 0xFF ; i++) {
		memory[i] = (byte *) malloc(65536);
		if (memory[i] == NULL) {
			printf("Failed\n");
			return 1;
		}
	}
	printf("Done\n");

	printf("    - Allocating 128 KB slow RAM (2 banks): ");
	for (i = 0xE0 ; i <= 0xE1 ; i++) {
		memory[i] = (byte *) malloc(65536);
		if (memory[i] == NULL) {
			printf("Failed\n");
			return 1;
		}
	}
	printf("Done\n");

	printf("    - Allocating %d MB fast RAM (%d banks): ",mem_ramsize,banks);
	for (i = 0 ; i < banks ; i++) {
		memory[i] = (byte *) malloc(65536);
		if (memory[i] == NULL) {
			printf("Failed\n");
			return 1;
		}
	}
	printf("Done\n");

	printf("    - Loading ROM from file \"%s\": ",ROM_FILE);
	if ((fp = fopen(ROM_FILE,"r")) == NULL) {
		printf("Failed\n");
		return 2;
	}
	if (fread(memory[0xFC],1,65536,fp) != 65536) {
		printf("Failed\n");
		return 2;
	}
	if (fread(memory[0xFD],1,65536,fp) != 65536) {
		printf("Failed\n");
		return 2;
	}
	if (fread(memory[0xFE],1,65536,fp) != 65536) {
		printf("Failed\n");
		return 2;
	}
	if (fread(memory[0xFF],1,65536,fp) != 65536) {
		printf("Failed\n");
		return 2;
	}
	fclose(fp);
	printf("Done\n");

	return 0;
}

void MEM_update()
{
}

void MEM_reset()
{
	int	i;

	/* By default, initialize all special memory to	*/
	/* trap all accesses to the default fault	*/
	/* handlers. This is just to be on the safe	*/
	/* side...					*/

	for (i = 0x00 ; i <= 0xFF ; i++) {
		bank00[i].readPtr = bank00[i].writePtr = NULL;
		bank00[i].readOffset = bank00[i].writeOffset = 0;
		bank00[i].readFaultHandler = MEM_defaultReadFaultHandler;
		bank00[i].writeFaultHandler = MEM_defaultWriteFaultHandler;
		bank01[i].readPtr = bank01[i].writePtr = NULL;
		bank01[i].readOffset = bank01[i].writeOffset = 0;
		bank01[i].readFaultHandler = MEM_defaultReadFaultHandler;
		bank01[i].writeFaultHandler = MEM_defaultWriteFaultHandler;
		bankE0[i].readPtr = bankE0[i].writePtr = NULL;
		bankE0[i].readOffset = bankE0[i].writeOffset = 0;
		bankE0[i].readFaultHandler = MEM_defaultReadFaultHandler;
		bankE0[i].writeFaultHandler = MEM_defaultWriteFaultHandler;
		bankE1[i].readPtr = bankE1[i].writePtr = NULL;
		bankE1[i].readOffset = bankE1[i].writeOffset = 0;
		bankE1[i].readFaultHandler = MEM_defaultReadFaultHandler;
		bankE1[i].writeFaultHandler = MEM_defaultWriteFaultHandler;
	}

	/* Most of banks $E0 & $E1 never change, so we	*/
	/* will initialize those maps here. 		*/

	for (i = 0x00 ; i <= 0xBF ; i++) {
		bankE0[i].readPtr = memory[0xE0] + (i << 8);
		bankE0[i].writePtr = bankE0[i].readPtr;
		bankE1[i].readPtr = memory[0xE1] + (i << 8);
		bankE1[i].writePtr = bankE1[i].readPtr;
	}

	/* Now just set all of our soft switches to a	*/
	/* sensible power-up state, and build the maps.	*/

	mem_80store = 0;
	mem_auxrd = 0;
	mem_auxwrt = 0;
	mem_altzp = 0;

	mem_lcbank2 = 0;
	mem_lcread = 0;
	mem_lcwrite = 0;
	mem_lcsecond = 0;

	mem_shadow_text = 1;
	mem_shadow_hires1 = 1;
	mem_shadow_hires2 = 1;
	mem_shadow_super = 0;
	mem_shadow_aux = 1;
	mem_shadow_lc = 1;

	for (i = 0 ; i < 5 ; i++) mem_slot_reg[i] = 0;
	mem_slot_reg[5] = 1;
	mem_slot_reg[6] = 1;
	mem_slot_reg[7] = 1;

	MEM_buildMap();
}

void MEM_shutdown()
{
	printf("\nShutting down emulator memory\n");
}

/* OK, this code isn't too elegant (nor is it very fast), but it gets	*/
/* the job done. It just goes to show how COMPLICATED the IIGS memory	*/
/* map really is...between the funky IIe modes and the shadow modes,	*/
/* it's a real mess in there.						*/

void MEM_buildMap()
{
	int	i,rdflag,wrflag;

	for (i = 0x00 ; i <= 0x01 ; i++) {
		bank00[i].readPtr = memory[mem_altzp] + (i << 8);
		bank00[i].writePtr = memory[mem_altzp] + (i << 8);
		bank00[i].readOffset = bank00[i].writeOffset = mem_altzp << 16;
		bank01[i].readPtr = memory[0x01] + (i << 8);
		bank01[i].writePtr = bank01[i].readPtr;
	}
	for (i = 0x02 ; i <= 0x03 ; i++) {
		bank00[i].readPtr = memory[mem_auxrd] + (i << 8);
		bank00[i].writePtr = memory[mem_auxwrt] + (i << 8);
		bank00[i].readOffset = mem_auxrd << 16;
		bank00[i].writeOffset = mem_auxwrt << 16;
		bank01[i].readPtr = memory[0x01] + (i << 8);
		bank01[i].writePtr = bank01[i].readPtr;
	}
	if (mem_80store) {
		rdflag = vid_page2;
		wrflag = vid_page2;
	} else {
		rdflag = mem_auxrd;
		wrflag = mem_auxwrt;
	}
	for (i = 0x04 ; i <= 0x07 ; i++) {
		bank00[i].readPtr = memory[rdflag] + (i << 8);
		bank00[i].readOffset = rdflag << 16;
		bank00[i].writeOffset = wrflag << 16;
		bank01[i].readPtr = memory[0x01] + (i << 8);
		if (mem_shadow_text) {
			bank00[i].writePtr = NULL;
			bank01[i].writePtr = NULL;
		} else {
			bank00[i].writePtr = memory[wrflag] + (i << 8);
			bank01[i].writePtr = bank01[i].readPtr;
		}
	}
	for (i = 0x08 ; i <= 0x0B ; i++) {
		bank00[i].readPtr = memory[mem_auxrd] + (i << 8);
		bank00[i].writePtr = memory[mem_auxwrt] + (i << 8);
		bank00[i].readOffset = mem_auxrd << 16;
		bank00[i].writeOffset = mem_auxwrt << 16;
		bank01[i].readPtr = memory[0x01] + (i << 8);
		bank01[i].writePtr = bank01[i].readPtr;
	}
	for (i = 0x0C ; i <= 0x1F ; i++) {
		bank00[i].readPtr = memory[mem_auxrd] + (i << 8);
		bank00[i].writePtr = memory[mem_auxwrt] + (i << 8);
		bank00[i].readOffset = mem_auxrd << 16;
		bank00[i].writeOffset = mem_auxwrt << 16;
		bank01[i].readPtr = memory[0x01] + (i << 8);
		bank01[i].writePtr = bank01[i].readPtr;
	}
	if (mem_80store) {
		rdflag = vid_page2;
		wrflag = vid_page2;
	} else {
		rdflag = mem_auxrd;
		wrflag = mem_auxwrt;
	}
	for (i = 0x20 ; i <= 0x3F ; i++) {
		bank00[i].readPtr = memory[rdflag] + (i << 8);
		bank00[i].writePtr = (mem_shadow_hires1? NULL : memory[wrflag] + (i << 8));
		bank00[i].readOffset = rdflag << 16;
		bank00[i].writeOffset = wrflag << 16;
		bank01[i].readPtr = memory[0x01] + (i << 8);
		bank01[i].writePtr = (((mem_shadow_hires1 && mem_shadow_aux) || mem_shadow_super)? NULL : bank01[i].readPtr);
	}
	for (i = 0x40 ; i <= 0x5F ; i++) {
		bank00[i].readPtr = memory[mem_auxrd] + (i << 8);
		bank00[i].writePtr = (mem_shadow_hires2? NULL : memory[mem_auxwrt] + (i << 8));
		bank00[i].readOffset = mem_auxrd << 16;
		bank00[i].writeOffset = mem_auxwrt << 16;
		bank01[i].readPtr = memory[0x01] + (i << 8);
		bank01[i].writePtr = (((mem_shadow_hires2 && mem_shadow_aux) || mem_shadow_super)? NULL : bank01[i].readPtr);
	}
	for (i = 0x60 ; i <= 0x9F ; i++) {
		bank00[i].readPtr = memory[mem_auxrd] + (i << 8);
		bank00[i].writePtr = memory[mem_auxwrt] + (i << 8);
		bank00[i].readOffset = mem_auxrd << 16;
		bank00[i].writeOffset = mem_auxwrt << 16;
		bank01[i].readPtr = memory[0x01] + (i << 8);
		bank01[i].writePtr = (mem_shadow_super? NULL : bank01[i].readPtr);
	}
	for (i = 0xA0 ; i <= 0xBF ; i++) {
		bank00[i].readPtr = memory[mem_auxrd] + (i << 8);
		bank00[i].writePtr = memory[mem_auxwrt] + (i << 8);
		bank00[i].readOffset = mem_auxrd << 16;
		bank00[i].writeOffset = mem_auxwrt << 16;
		bank01[i].readPtr = memory[0x01] + (i << 8);
		bank01[i].writePtr = bank01[i].readPtr;
	}
	if (mem_shadow_lc) {
		MEM_buildLanguageCard(mem_altzp,bank00);
		MEM_buildLanguageCard(0x01,bank01);
		rdflag = mem_altzp << 16;
		for (i = 0xD0 ; i <= 0xFF ; i++) {
			bank00[i].readOffset = rdflag;
			bank00[i].writeOffset = rdflag;
			bank01[i].readOffset = 0;
			bank01[i].writeOffset = 0;
			bankE0[i].readOffset = 0;
			bankE0[i].writeOffset = 0;
			bankE1[i].readOffset = 0;
			bankE1[i].writeOffset = 0;
		}
	} else {
		for (i = 0xC0 ; i <= 0xFF ; i++) {
			bank00[i].readPtr = memory[mem_auxrd] + (i << 8);
			bank00[i].writePtr = memory[mem_auxwrt] + (i << 8);
			bank00[i].readOffset = mem_auxrd << 16;
			bank00[i].writeOffset = mem_auxwrt << 16;
			bank01[i].readPtr = memory[0x01] + (i << 8);
			bank01[i].writePtr = bank01[i].readPtr;
		}
	}
	MEM_buildLanguageCard(0xE0,bankE0);
	MEM_buildLanguageCard(0xE1,bankE1);
}

/* This routine builds a language card in the given bank.  This	*/
/* means building all mapping info from $C000 up to $FFFF.	*/

void MEM_buildLanguageCard(int bank,mem_pagestruct *page)
{
	int	i,offset;

	page[0xC0].readPtr = NULL;
	page[0xC0].writePtr = NULL;
	page[0xC0].readFaultHandler = MEM_ioReadFaultHandler;
	page[0xC0].writeFaultHandler = MEM_ioWriteFaultHandler;
	for (i = 0xC1 ; i<= 0xCF ; i++) {
		page[i].readPtr = memory[0xFF] + (i << 8);
		page[i].writePtr = NULL;
	}

	page[0xC5].readPtr = dsk_slot5_rom;
	page[0xC6].readPtr = NULL;
	page[0xC7].readPtr = dsk_slot7_rom;

	offset = mem_lcbank2? 0x1000 : 0x0000;
	for (i = 0xD0 ; i <= 0xDF ; i++) {
		if (mem_lcread) {
			page[i].readPtr = memory[bank] + (i << 8) - offset;
		} else {
			page[i].readPtr = memory[0xFF] + (i << 8);
		}
		page[i].writePtr = (mem_lcwrite? memory[bank] + (i << 8) - offset : NULL);
	}
	for (i = 0xE0 ; i <= 0XFF ; i++) {
		if (mem_lcread) {
			page[i].readPtr = memory[bank] + (i << 8);
		} else {
			page[i].readPtr = memory[0xFF] + (i << 8);
		}
		page[i].writePtr = (mem_lcwrite? memory[bank] + (i << 8) : NULL);
	}
}
